/* wordtable.c -- Copyright 1989, 1993 Liam R. Quin.  All Rights Reserved.
 * This code is NOT in the public domain.
 * See the file ../COPYRIGHT for full details.
 */

/* Symbol Table Interface to text retrieval database.
 * Handles both the internal and external indexes.
 *
 * This originally used a linked list.  Converting to a hash table reduced
 * the time to index comp.os.vms from nearly an hour to one and a half
 * minutes...
 *
 * Liam Quin, 1989
 */

/* 
 * $Id: wordtable.c,v 2.23 93/12/08 23:16:04 lee Exp $
 */

#ifndef lint
 static char *Rcs = "$Id: wordtable.c,v 2.23 93/12/08 23:16:04 lee Exp $";
#endif

#include <stdio.h>
#include <malloc.h>
#include <ctype.h>
#include <sys/types.h>
#include <fcntl.h> /* for O_RDWR wtc */

#include "globals.h" /* defines and declarations for database filenames */
#include "error.h"

#include "smalldb.h"
#include "fileinfo.h"
#include "wordinfo.h"
#include "pblock.h"
#include "wordrules.h"
#include "emalloc.h"
#include "addfile.h"

extern t_WID GetNextWID();

static void NewEntry();
static void UpdateEntry();

#ifndef MaxWordPlacesInAWordBlock
# define MaxWordPlacesInAWordBlock ((WIDBLOCKSIZE-(MinWordLength+2)/3))
#endif

#ifndef HASHSIZ
# define HASHSIZ 32768 /* MUST be a power of two */
#endif /*!HASHSIZ*/

#ifndef MAXWORDSINCACHE
# define MAXWORDSINCACHE  (HASHSIZ * 10)
#endif

int MaxWordsInCache = MAXWORDSINCACHE;

extern int AsciiTrace;

/** System calls and library functions used in this file: **/

/** Lqtext calls */
extern unsigned int Putpblock();
extern void DeleteWordPlaces();
extern t_WordInfo *MakeWordInfo();
extern void SlayWordInfo();

/** calls within this file **/
void DumpCache(); /* see ../h/addfile.h */
static void FastDump();
static void DumpOldest();
static void DumpFrequent();
/** System calls: */

/** Library Functions: */
extern char *strncpy();
extern void perror();
extern void exit();

#ifdef ASCIITRACE
extern void fprintWordInfo();
#endif
/**/

extern char *progname;
static int HashSize = HASHSIZ; /* MUST be a power of two */
int
SetHashSize(n)
    int n;
{
    return HashSize = n;
}

#define NPLACES 4
#define NPLACESBIGINCR 16
#define NPLACESHUGEINCR 128
/* This is small to optimise the common case -- by far the majority of
 * words are used less than 10 times.  In the cases where we've gone
 * wrong, well, there'll be a few thousand.  We add slowly until we
 * get to NPLACE * 3, and then we go up in NPLACESBIGINCR lumps.
 */

typedef struct s_HashEl {
    char *Word;
    t_WID WID;
    int PlacesUsed;
    int PlacesAllocated;
    t_WordPlace *Places;
} t_HashEl;

static t_HashEl **SymbolTable;
static int LastEl = 0;
static int WordsInCache = 0;
static int DumpThresh = DUMP_FAST_THRESH;
static int SlotsUsed = 0;

void SetDumpThresh(Thresh)
    int Thresh;
{
    /* Set the threshhold for fast dumping.
     * If a word has less than this many occurrences in the cache, it gets
     * written out.  -1 disables this feature, and 0 uses the default.
     */

    DumpThresh = Thresh;

    if (!DumpThresh) {
	DumpThresh = DUMP_FAST_THRESH;
    }
}

static void
InitHash()
{
    long tmp = 1;
    while (tmp < HashSize) {
	tmp <<= 1;
    }
    HashSize = tmp;

    if (HashSize < 1) {
	Error(E_FATAL, "InitHash: hash size (%d/%d) is too small!\n",
				HashSize, MaxWordsInCache);
    }

    SymbolTable = (t_HashEl **) ecalloc(
	"AddWord symbol table",
	HashSize,
	sizeof(t_HashEl *)
    );
    LastEl = HashSize; /* Used as a sentinel */
#ifdef ASCIITRACE
    if (AsciiTrace > 2) {
	fprintf(stderr, "%s: allocated %ld hash slots for up to %ld words\n",
	    progname,
	    HashSize,
	    MaxWordsInCache
	);
    }
#endif
}

static t_HashEl ZeroEl = {
    0,
};

#define SetElEmpty(El) \
    (*(El)) = ZeroEl; /* structure assignment */ \
    (El)->PlacesAllocated = NPLACES; \
    (El)->PlacesUsed = 0; \
    (El)->Places = (t_WordPlace *) emalloc( \
	"wordtable::SetElEmpty.Places", sizeof(t_WordPlace) * NPLACES);

#ifndef Hash
INLINE
int
Hash(Word, Length)
    register char *Word;
    register int Length;
{
    register unsigned long n = 0;

#ifndef NODUFF /* clever stuff for speedup... dmr-approved!... */

#define HASHC	n = *Word++ + 65599 * n

    if (Length > 0) {
	register int loop = (Length + 8 - 1) >> 3;

	switch(Length & (8 - 1)) {
	case 0:	do {
		HASHC;	case 7:	HASHC;
	case 6:	HASHC;	case 5:	HASHC;
	case 4:	HASHC;	case 3:	HASHC;
	case 2:	HASHC;	case 1:	HASHC;
		} while (--loop);
	}

    }
#else /* NODUFF */
    while (Length--)
	n = *Word++ + 65599 * n;
#endif /* NODUFF */
    /**
    return n & (HashSize - 1);
    **/
    return n % HashSize;
}
#endif /* Hash */

static char FirstTimeRound = 1;

static t_FID *FIDsToDelete;
static int FIDsToDeleteCount = 0;
static int FIDSAllocated = 0;

void
AddFIDToDelete(FID)
    t_FID FID;
{
    int i;

    for (i = 0; i < FIDsToDeleteCount; i++) {
	if (FIDsToDelete[i] == FID) {
	    return; /* already there */
	}
    }


    if (FIDsToDeleteCount >= FIDSAllocated) {
	FIDsToDelete = (t_FID *) erealloc(
	    FIDsToDelete,
	    FIDsToDeleteCount * sizeof(int)
	);
	FIDsToDelete[FIDsToDeleteCount] = FID;
	++FIDsToDeleteCount;
    }
}

static int
FIDIsWanted(WID, WordPlace)
    t_WID WID;
    t_WordPlace *WordPlace;
{
    register t_FID f;
    static int Last = 0;
    register int j;

    if (FIDsToDelete == (t_FID *) 0) {
	return 1;
    }

    f = WordPlace->FID;

    if (FIDsToDelete[Last] == f) return 0;

    for (j = Last + 1; j != Last; j++) {
	if (j == FIDsToDeleteCount) {
	    j = 0;
	}
	if (FIDsToDelete[j] == f) {
	    Last = j;
	    return 0;
	}
    }
    return 1;
}

static int Deleting = 0;
static int WordsToDelete = 0;

void
DelWord(WordInfo, FID)
    t_WordPlace *WordInfo;
    t_FID FID;
{
    void AddWord();

    ++Deleting;
    AddWord(WordInfo);
    --Deleting;
}

static t_FID CurrentFID = 0;

void
AddWord(WordInfo)
    t_WordInfo *WordInfo;
{
    register t_HashEl *HashEl;
    int Slot, FirstSlot;
    static t_FID LastFID = 0;
    int ProbeCount = 0;

    if (!WordInfo || !WordInfo->Word || !WordInfo->Word[0]) {
	Error(E_WARN, "Null word in AddWord(0x%x)", WordInfo);
	return;
    }

#ifdef ASCIITRACE
    if (AsciiTrace & 64) {
	fprintf(stderr,
	    " <%s, b %d, w %d> ",
	    WordInfo->Word,
	    WordInfo->WordPlace.BlockInFile,
	    WordInfo->WordPlace.WordInBlock
	);
    }
#endif

    if (!LastEl) {
	InitHash();
	if (FirstTimeRound) {
	    /* Special check to save looking up the WIDs first time round */
	    t_WID W = GetMaxWID();

	    if (W == 0L) {
		if (AsciiTrace > 1) {
		    fprintf(stderr, "\n%s: first ever run, allocating WIDs.\n",
			    progname);
		}
		FirstTimeRound = 1; /* actually it's already 1 here */
	    } else {
		FirstTimeRound = 0;
	    }
	}
    } else {
	if (WordInfo->WordPlace.FID != LastFID) {
	    if (WordInfo->WordPlace.FID == 0L) {
		Error(E_BUG, "AddWord: FID 0 for \"%s\"", WordInfo->Word);
	    } else if (LastFID != 0) {
		/* quick opportunity to get rid of words that
		 * were only in the previous file -- this takes advantage
		 * of the fact that vocabulary is document-specific,
		 * i.e. you tend to get words used in one file and not
		 * another.
		 */
		CurrentFID = WordInfo->FID;
		if (MUST_FILE_DUMP(
			    WordsInCache,
			    MaxWordsInCache,
			    SlotsUsed,
			    HashSize
		)) {
		    DumpCache(DUMP_CACHE|DUMP_FAST);
		}
	    }
	    LastFID = WordInfo->WordPlace.FID;
	}
    }


    FirstSlot = Slot = Hash(WordInfo->Word, WordInfo->Length);

    for (ProbeCount = 0; ProbeCount < MAXPROBES; ProbeCount++) {
	if (SymbolTable[Slot] == (t_HashEl *) NULL) {
	    extern char *strcpy();
	    extern t_WID Word2WID();

	    /* make a new element */
	    HashEl = SymbolTable[Slot] = (t_HashEl *) emalloc(
		"wordtable::AddWord.HashEl",
		sizeof(t_HashEl)
	    );
	    SetElEmpty(HashEl);
	    HashEl->Word = emalloc(
		"wordtable::AddWord.Word",
		WordInfo->Length + 1
	    );
	    (void) strcpy(HashEl->Word, WordInfo->Word);
	    if (FirstTimeRound) {
		/* No point looking for the WID, it won't be there;
		 * code in DumpCache() will assign a new one.  Word2WID()
		 * will always return zero on a new index.
		 */
		HashEl->WID = 0;
	    } else {
		HashEl->WID = Word2WID(HashEl->Word, WordInfo->Length);
	    }
	    ++SlotsUsed;
	    break;
	} else if (STREQ(SymbolTable[Slot]->Word, WordInfo->Word)) {
	    HashEl = SymbolTable[Slot];
	    if (Slot != FirstSlot) {
		if (
		    HashEl->PlacesUsed >
		    SymbolTable[FirstSlot]->PlacesUsed + 2
		) {
		    /* Swap so we find it first time next time round.
		     * this is inefficient if the words occur approximately
		     * alternately with equal frequency, but the most frequent
		     * few words accounting for 75% of the data don't do that.
		     * The +2 requirement stops this happening every time
		     * round even in the worst case.
		     */
		    t_HashEl *tmp = SymbolTable[FirstSlot];
		    HashEl = SymbolTable[FirstSlot] = SymbolTable[Slot];
		    SymbolTable[Slot] = tmp;
		    Slot = FirstSlot;
		}
	    }
	    break;
	}

	if ((Slot += 17) >= HashSize) Slot %= HashSize;
	/** Note: hashsize is a power of 2, so we can step through it
	 ** in any prime number, but doing so in little chunks is best
	 ** because we don't hit so many pages, but at the same time
	 ** step over clusters.  I hope.
	 **/

	/** If the hash table is full, no point looking further: **/
	if (ProbeCount + 1 >= MAXPROBES) {
#ifdef ASCIITRACE
	    if (AsciiTrace > 1) {
		fprintf(stderr,
		    "(%d probes didn't find a slot for %s)\n",
		    ProbeCount, WordInfo->Word
		);
	    }
#endif
	    CurrentFID = WordInfo->FID;
	    DumpCache(DUMP_CACHE|DUMP_FAST);
	    Slot = FirstSlot;
	    ProbeCount = 0;
	}
    }

    /* check to see if there is already an occurrence of this word
     * for this file:
     */

    if (Deleting) {
	register int i;

	for (i = HashEl->PlacesUsed - 1; i >= 0; i--) {
	    if (HashEl->Places[i].FID == WordInfo->WordPlace.FID) {
		return;
	    }
	}

	++WordsToDelete;
    }

    /* If we get here, all we need to do is add the WordPlace */

    if (HashEl->PlacesAllocated - HashEl->PlacesUsed <= 0) {
	if (HashEl->PlacesAllocated <= NPLACES * 3) {
	    HashEl->PlacesAllocated += NPLACES;
	} else if (HashEl->PlacesAllocated <= NPLACESBIGINCR) {
	    HashEl->PlacesAllocated += NPLACESBIGINCR;
	} else {
	    HashEl->PlacesAllocated += NPLACESHUGEINCR;
	}
	HashEl->Places = (t_WordPlace *) erealloc(
	    (char *) HashEl->Places,
	    sizeof(t_WordPlace) * HashEl->PlacesAllocated
	);
    }

#ifdef ASCIITRACE
    if (AsciiTrace > 3) {
	if (HashEl->PlacesUsed) {
	    t_WordPlace *wp = &HashEl->Places[HashEl->PlacesUsed - 1];

	    if (wp->FID > WordInfo->WordPlace.FID) {
		Error(E_BUG, "AddWord: %s: place %d: FID %ld > %ld", 
		    HashEl->Word,
		    HashEl->PlacesUsed + 1,
		    wp->FID,
		    WordInfo->WordPlace.FID
		);
	    } else if (wp->FID == WordInfo->WordPlace.FID) {
		if (wp->BlockInFile > WordInfo->WordPlace.BlockInFile) {
		    Error(E_BUG, "AddWord: %s: place %d: FID %ld: Blk %ld > %ld",
			HashEl->Word,
			HashEl->PlacesUsed + 1,
			wp->FID,
			wp->BlockInFile,
			WordInfo->WordPlace.BlockInFile
		    );
		} else if (wp->BlockInFile == WordInfo->WordPlace.BlockInFile &&
			    wp->WordInBlock >= WordInfo->WordPlace.WordInBlock) {
		    Error(E_BUG,
			"AddWord: %s: place %d: FID %ld: Blk %ld: WIB %d > %d",
			HashEl->Word,
			HashEl->PlacesUsed + 1,
			wp->FID,
			wp->BlockInFile,
			wp->WordInBlock,
			WordInfo->WordPlace.WordInBlock
		    );
		}
	    }
	}
    }
#endif /*ASCIITRACE*/
    HashEl->Places[HashEl->PlacesUsed++] = WordInfo->WordPlace;

    WordsInCache++;

#ifdef ASCIITRACE
    if (AsciiTrace > 100) {
	fprintf(stderr, "Slot %d Word %s len %d places %d\n",
		Slot, SymbolTable[Slot]->Word,
		WordInfo->Length, SymbolTable[Slot]->PlacesUsed);
    }
#endif
    return;
}

static void
DumpOneEntry(HashEl, WordLength)
    t_HashEl *HashEl;
    int WordLength; /* avoids extra calls to strlen() */
{
    /* We are going to make a new index entry for the word.
     * There are two cases -- depending on whether the word
     * is already indexed or not.
     * In the former case we must merge the new information.
     * In the latter case we don't have to read the old info,
     * but we must make a new entry in the WID Index.
     */

    WordsInCache -= HashEl->PlacesUsed;

    if (HashEl->WID == 0) {
	NewEntry(HashEl, WordLength);
    } else {
	UpdateEntry(HashEl, WordLength);
    }

    /* Reclaim storage */
    /* Note: no point testing CallFree, as it's always set if
     * we're in here.  There's no point using the partial cache
     * dump routine if you're not going to free the slots, as then
     * you get left with a corrupt symbol table!  It's only useful
     * for the final total cache dump.
     */
    efree(HashEl->Word);
    efree((char *) HashEl->Places);
    efree((char *) HashEl);
}

#define TOPCOUNT 25

static void
DumpFrequent(CallFree)
    int CallFree;
{
    extern void FlushCache();
    register int i;
    register t_HashEl *HashEl;
    int Progress = 0;
    int TopFew[TOPCOUNT];
    int InTopFew = 0;

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "\n[frequent dump: %d/%d words, %d/%d slots ",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    for (i = 0; i != LastEl; i++) {
	long PlaceCount;

	if (!SymbolTable[i]) continue;

	HashEl = SymbolTable[i];

	if (InTopFew < TOPCOUNT) {
	    TopFew[InTopFew++] = i;
	} else {
	    register int j;

	    PlaceCount = HashEl->PlacesUsed;

	    for (j = 0; j < TOPCOUNT; j++) {
		if (SymbolTable[TopFew[j]]->PlacesUsed < PlaceCount) {
		    TopFew[j] = i;
		    /* Since this is the smallest entry in the first
		     * n places,
		     * we'll swap it with the previous one.
		     * In this way, we hope that smaller entries will
		     * tend to move a litle nearer the front, and
		     * we'll run faster.
		     * Also, this will help make the routine work
		     * properly...
		     */
		    if (j > 1) {
			int tmp = TopFew[j];
			TopFew[j] = TopFew[j - 1];
			TopFew[j - 1] = tmp;
		    }
		    break;
		}
	    }
	}
    }

    for (i = 0; i < InTopFew; i++) {
#ifdef ASCIITRACE
	if (AsciiTrace > 2) {
	    fprintf(stderr,
		"dump word %s, occ %ld\n",
		SymbolTable[TopFew[i]]->Word,
		SymbolTable[TopFew[i]]->PlacesUsed
	    );
	}
#endif
	DumpOneEntry(
	    SymbolTable[TopFew[i]],
	    strlen(SymbolTable[TopFew[i]]->Word)
	);
	SymbolTable[TopFew[i]] = 0;
	--SlotsUsed;
    }

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "frequent words... cache now %d/%d, slots %d/%d]\n",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    FirstTimeRound = 0;
    FlushCache(0); /* the number is non-zero if we only need to clear 1 slot */
}

static void
DumpOldest(CallFree)
    int CallFree;
{
    extern void FlushCache();
    static t_FID LastCalledFor = 0;
    register int i;
    register t_HashEl *HashEl;
    int Progress = 0;

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "\n[oldest dump: %d/%d words, %d/%d slots ",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    if (LastCalledFor == CurrentFID) {
	if (AsciiTrace > 1) {
	    fprintf(stderr, "-- but still in same file\n");
	}
	return;
    }

    for (i = 0; i != LastEl; i++) {

	if (!SymbolTable[i]) continue;

	HashEl = SymbolTable[i];

	if (HashEl->Places[HashEl->PlacesUsed - 1].FID != CurrentFID) {
	    DumpOneEntry(HashEl, strlen(HashEl->Word));
	    SymbolTable[i] = 0;
	    --SlotsUsed;
	}

	if (AsciiTrace > 1) {
	    if (i >= Progress * (HashSize / 16)) {
		fputc("1234567890abcdef?!"[Progress], stderr);
		fflush(stderr);
		++Progress;
	    }
	}
    }

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "-- cache now %d/%d, slots %d/%d]\n",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    FirstTimeRound = 0;
    FlushCache(0); /* the number is non-zero if we only need to clear 1 slot */

    if (MUST_FILE_DUMP_MORE(
	WordsInCache,
	MaxWordsInCache,
	SlotsUsed,
	HashSize
    )) {
	DumpCache(DUMP_CACHE|DUMP_FAST);
    }
}

static void
FastDump(CallFree)
    int CallFree;
{
    extern void FlushCache();
    register int i;
    register t_HashEl *HashEl;
    int Progress = 0;

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "\n[fast dump nocc < %d, from %d/%d words, %d/%d slots ",
	    DumpThresh,
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    for (i = 0; i != LastEl; i++) {
	int ForceDump;
	int len;

	if (!SymbolTable[i]) continue;

	HashEl = SymbolTable[i];
	len = strlen(HashEl->Word);
	ForceDump = 0;

	if (HashEl->PlacesUsed > DumpThresh) {
	    /* We're not going to dump this entry.
	     * If there was a collision (two or more entries have the same
	     * hash code), we mustn't simply delete the entry at the hashed
	     * slot, because then we would no longer find this one.
	     */
	    int Slot = Hash(HashEl->Word, len);

	    if (Slot != i) {
		/* It's in the wrong place...
		 * See if we can put it back:
		 */
		if (SymbolTable[Slot] == 0) {
		    /* In this case we can simply put it in the right place*/
		    SymbolTable[Slot] = HashEl;
		    SymbolTable[i] = 0;
		} else if (SymbolTable[Slot]->PlacesUsed <= DumpThresh) {
		    /* In this case the entry at the other location is
		     * going to be dumped, so we can swap them even though
		     * that puts the other entry temporarily in the wrong
		     * place.  We're about to delete it anyway.
		     */
		    register t_HashEl *tmp = SymbolTable[Slot];

		    SymbolTable[Slot] = HashEl;
		    HashEl = SymbolTable[i] = tmp;
		} else {
		    /* In this case, it's in the wrong place, and the
		     * entry in the other slot isn't going to be deleted.
		     * But it might be moved.  This (we hope) is rare
		     * enough that we handle it by giving up and
		     * dumping it anyway:
		     */
		    ForceDump = 1;
		}

		if (!SymbolTable[i]) {
		    continue; /* dealt with slot i. */
		}
	    }
	}

	if (HashEl->PlacesUsed <= DumpThresh || ForceDump) {
	    DumpOneEntry(HashEl, len);
	    SymbolTable[i] = 0;
	    --SlotsUsed;
	}

	if (AsciiTrace > 1) {
	    if (i >= Progress * (HashSize / 16)) {
		fputc("1234567890abcdef?!"[Progress], stderr);
		fflush(stderr);
		++Progress;
	    }
	}
    }

    if (AsciiTrace > 1) {
	fprintf(stderr,
	    "-- cache now %d/%d, slots %d/%d]\n",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    FirstTimeRound = 0;
    FlushCache(0); /* the number is non-zero if we only need to clear 1 slot */
}


void
DumpCache(CallFree)
    int CallFree;
{
    extern void FlushCache();
    register int i;
    register t_HashEl *HashEl;
    int Progress = 0;

    if (AsciiTrace > 1) {
	fprintf(stderr, "\nCache write from %d/%d words, %d/%d slots\n",
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	);
    }

    if (DumpThresh == -1) {
	CallFree |= DUMP_SYNC;
    } else if ((CallFree & DUMP_SYNC) == 0 || (CallFree & DUMP_FAST) != 0) {
	int SaveThresh = DumpThresh;
	int Multiplier = 2;

	/* ensure that we always do something useful:*/

	if (SlotsUsed * 3 > HashSize) {
	    DumpOldest(CallFree);
	}
	DumpFrequent(CallFree);

	/* now try and get below the dump threshhold: */

	while (MUST_DUMP_MORE(
	    WordsInCache,
	    MaxWordsInCache,
	    SlotsUsed,
	    HashSize
	)) {
	    if (SlotsUsed * 3 > HashSize) {
		FastDump(CallFree);
		DumpThresh *= Multiplier;
		++Multiplier;
	    }
	    DumpFrequent(CallFree);
	}
	DumpThresh = SaveThresh;
	return;
    }

    for (i = 0; i != LastEl; i++) {
	if (SymbolTable[i]) {
	    DumpOneEntry(SymbolTable[i], strlen(SymbolTable[i]->Word));

	    SymbolTable[i] = 0;
	    --SlotsUsed;
	}

	if (AsciiTrace > 1) {
	    if (i >= Progress * (HashSize / 16)) {
		fputc("1234567890ABCDEF?!"[Progress], stderr);
		++Progress;
	    }
	}

	if (SlotsUsed == 0 || WordsInCache <= 0) break;
    }

#ifdef ASCIITRACE
    if (WordsInCache || SlotsUsed) {
	Error(E_BUG, "Cache, %d words left, %d slots [%s:%d]",
	    WordsInCache, SlotsUsed, __FILE__, __LINE__
	);
    }
#endif
    WordsInCache = 0;
    FlushCache(0); /* the number is non-zero if we only need to clear 1 slot */

    if (AsciiTrace) {
	fprintf(stderr, "\nCache now empty\n");
    }
    FirstTimeRound = 0;
	/* the first time round we don't bother looking up words to see
	 * if they already have a WID.
	 */

    WriteCurrentMaxWID();
    cleanupdb(); /* allow other processes a chance to do some reading */
    lqWriteAccess();
}

static void
NewEntry(HashEl, Length)
    t_HashEl *HashEl;
    int Length;
{
    t_pblock *pblock = 0;
    register int i, k;
    t_WordInfo *WordInfo;
    /* TODO: add MightNeedToSort check */

    /** make a WIDIndex entry and mark it as invalid (NOTDONE) */

    /* In order to do this, we must make a "pblock", a structure that
     * reflects the physical database.  This is fairly low-level stuff
     * for efficiency's sake...
     */

    /* allocate a pblock structure.  These are rather devious things, a
     * structure with an array tacked onto the end.
     */
    pblock = (t_pblock *) emalloc(
	"wordtable::NewEntry.pblock",
	sizeof(t_pblock) + HashEl->PlacesUsed * sizeof(t_WordPlace)
    );
    
    pblock->WID = HashEl->WID;
    pblock->ChainStart = 0L; /* address on disk -- not there yet, so 0! */
    pblock->NumberOfWordPlaces = HashEl->PlacesUsed;

    /* fill in the WordPlaces */
    for (i = k = 0; i < HashEl->PlacesUsed; i++) {
	if (FIDIsWanted(HashEl->WID, &HashEl->Places[i])) {
	    pblock->WordPlaces[k] = HashEl->Places[i]; /* struct copy */
	    ++k;
	}
    }

    /* if k is 0, they are all deleted, but that's unaccepatable.
     * help!
     */
    if (k == 0) {
#ifdef ASCIITRACE
	if (AsciiTrace > 4) {
	    fprintf(stderr, "no places to add for %s\n", HashEl->Word);
	}
#endif
	efree(pblock);
	return;
    }

    if (HashEl->WID == 0) {
	HashEl->WID = AllocateWID(HashEl->Word, Length, GetNextWID(0));
    }
 
    /* Now fill in enough of WordInfo to let us use the low-level routines: */
    WordInfo = MakeWordInfo(HashEl->WID, Length, HashEl->Word);
    WordInfo->Offset = 0L;

    WordInfo->NumberOfWordPlaces = pblock->NumberOfWordPlaces;

    /* First, let's make an index entry: */

    if (pblock->NumberOfWordPlaces <= MaxWordPlacesInAWordBlock) {
	(void) MkWIB(WordInfo, pblock);
    }

    /** write out the new entry */
    if (WordInfo->WordPlacesInHere != pblock->NumberOfWordPlaces) {
	/* In this case, it didn't all fit into the WID index block: */
	(void) Putpblock(WordInfo, pblock);
    }

    if (PutWordInfoIntoIndex(WordInfo, pblock->ChainStart) < 0) {
	Error(E_SYS|E_FATAL,
	    "NewEntry: Couldn't insert \"%s\" in database at 0x%lx",
			    WordInfo->Word, pblock->ChainStart);
    }

    /** reclaim storage */
    if (pblock) {
	(void) efree((char *) pblock);
    }
    SlayWordInfo(WordInfo);
}

static void
UpdateEntry(HashEl, Length)
    t_HashEl *HashEl;
    int Length;
{
    extern unsigned long AddWordPlaces();
    extern t_pblock *Getpblock();
    extern t_WordInfo *WID2WordInfo();

    register int i;
    t_pblock *pblock;
    t_WordInfo *WordInfo;
    int MightNeedToSort = 0;

#ifdef ASCIITRACE
    if (AsciiTrace >= 4) {
	fprintf(stderr, "UpdateEntry(%s/WID %ld, wordlen %d)\n",
			HashEl->Word, HashEl->WID, Length);
    }
#endif

    /** get the old entry */

    if (!HashEl->WID || !(WordInfo = WID2WordInfo(HashEl->WID))) {
	Error(E_BUG, "Word %s WID %ld went away!", HashEl->Word, HashEl->WID);
	NewEntry(HashEl, Length);
	return;
    }

#ifdef ASCIITRACE
    if (AsciiTrace & 32) {
	fprintWordInfo(stderr, WordInfo, "UpdateEntry");
    }
#endif

    if (WordInfo->Offset) {
	(void) AddWordPlaces(
	    HashEl->Places,
	    HashEl->WID,
	    WordInfo->Offset,
	    (unsigned long) HashEl->PlacesUsed
	);
	(void) SlayWordInfo(WordInfo);
	return;
    }

    if (WordInfo->WordPlacesInHere == WordInfo->NumberOfWordPlaces) {
	pblock = (t_pblock *) 0;
    } else {
	pblock = Getpblock(WordInfo);
    }

    if (pblock) {
	pblock = (t_pblock *) erealloc((char *) pblock, sizeof(t_pblock) +
	     (pblock->NumberOfWordPlaces +
				    HashEl->PlacesUsed) * sizeof(t_WordPlace));

    } else {
	pblock = (t_pblock *) emalloc(
	    "wordtable:NewEntry.pblock",
	    sizeof(t_pblock) +
	    (WordInfo->WordPlacesInHere + HashEl->PlacesUsed) *
							sizeof(t_WordPlace)
	);
	pblock->NumberOfWordPlaces = 0;
        if (WordInfo->WordPlacesInHere < WordInfo->NumberOfWordPlaces) {
	    extern t_WordPlace *GetWordPlacesWhere();

	    if (WordInfo->WordPlaceStart) {
		WordInfo->WordPlaces = GetWordPlacesWhere(
		    WordInfo->WID,
		    WordInfo->WordPlaceStart,
		    WIDBLOCKSIZE - (WordInfo->WordPlaceStart - WordInfo->DataBlock),
		    0L,
		    WordInfo->NumberOfWordPlaces,
		    FIDIsWanted
		);
	    }
	}

	/* Assert: the wordplaces in WordInfo are sorted */
	for (i = 0; i < WordInfo->NumberOfWordPlaces; i++) {
	    pblock->WordPlaces[pblock->NumberOfWordPlaces++] =
				WordInfo->WordPlaces[i]; /* structure copy */

	}
    }

    /* delete the old entry from disk */
    if (WordInfo->Offset) {
	/* Remove the old information from disk.
	 * This isn't as bad as it sounds, as it will be at the start
	 * of the freelist, so when we write it out again it will be
	 * in the buffer cache...  But it would still be faster to append.
	 */
	DeleteWordPlaces(WordInfo->Offset, WordInfo->WID);
    }

    pblock->WID = HashEl->WID;
    WordInfo->Offset = pblock->ChainStart = 0L; /* it's invalid now... */

    /* Merge the WordPlaces */

    /* Assert: we need only compare the last old entry and the
     * first new one to see if we might need a sort.  Note that
     * there must _be_ entries in pblock, as otherwise we'd have called
     * NewEntry() and not UpdateEntry().
     */

    if (pblock->WordPlaces[pblock->NumberOfWordPlaces - 1].FID >=
				HashEl->Places[0].FID) {
	MightNeedToSort = 1;
    }

    for (i = 0; i < HashEl->PlacesUsed; i++) {
	pblock->WordPlaces[pblock->NumberOfWordPlaces++] =
				HashEl->Places[i]; /* copy the struct: */
	/* TODO: call qcmp to check for sorting (actually only need to
	 * check the FIDs of the new entries)
	 */
    }
    
    if (MightNeedToSort) {
	extern void SortWordPlaces();

	SortWordPlaces(pblock->NumberOfWordPlaces, pblock->WordPlaces);
    }

    WordInfo->NumberOfWordPlaces = pblock->NumberOfWordPlaces;

    /* First, let's make an index entry: */
    if (pblock->NumberOfWordPlaces <= MaxWordPlacesInAWordBlock) {
	(void) MkWIB(WordInfo, pblock);
    }

    /** write out the new entry */
    if (WordInfo->WordPlacesInHere == pblock->NumberOfWordPlaces) {
	/* In this case it all fits into the WID index */
	pblock->ChainStart = 0L;
    } else {
	(void) Putpblock(WordInfo, pblock);
    }
    if (PutWordInfoIntoIndex(WordInfo, pblock->ChainStart) < 0) {
	Error(E_FATAL|E_SYS,
	    "UpdateEntry: Couldn't update \"%s\" in database at 0x%lx",
			    WordInfo->Word, pblock->ChainStart
	);
    }

    /** reclaim storage */
    if (pblock) {
	(void) efree((char *)pblock);
    }
    (void) SlayWordInfo(WordInfo);
}
